""" Controller for the `Simulator` class.
"""


class Engine:
    def __init__(self, smart_city):
        self.smart_city = smart_city
        self.scores = []

    def __str__(self):
        pass

    def light_up(self, lights_on):
        self.smart_city.lights_on()

    def heat_up(self, heats, temperatures):
        self.smart_city.heat_up(heats, temperatures)

    def buy_energies(self, amounts):
        self.smart_city.buy_energies(amounts)

    def step(self):
        step_scores = self.smart_city.step()
        self.scores.append(step_scores)



__name__ = '.'.join(__name__.split('/'))
__package__ = '.'.join('.'.join(__name__.split('/')).split('.')[:-1])

__author__ = "Adrien Pouyet (Ricocotam)"

from gym.envs.registration import register

register(
    id='smartcity-small',
    entry_point='gym_foo.envs:FooEnv',
)


import gym


class SmartCityEnv(gym.Env):
    default_json_path = "settings.json"

    def __init__(self):
        pass

    def change_settings(self, setting_file_path):
        pass

    def step(self, **kwargs):
        obs, reward, done, info = [None] * 4
        return obs, reward, done, info

    def seed(self, seed):
        pass

    def reset(self):
        pass

    def render(self):
        pass

    def close(self):
        pass



def estimate_beta_parameters(means, stds, low, high):
    loc = low
    scale = high

    # print("\t\t BEFORE", means, stds)
    means = (means - loc) / scale
    variances = (stds / scale) ** 2
    # print("\t\t AFTER", means, stds / scale)

    temp = means * (1 - means) / variances - 1
    α = means * temp
    β = (1 - means) * temp

    print(f"""
        Estimation of Beta Parameters
            α = {α}
            β = {β}
            loc = {loc}
            scale = {scale}

        Computed using :
            means : {means}
            vars : {variances}
            loc : {loc}
            scale : {scale}
    """)

    return α, β, loc, scale


class BetaDistributions(object):
    def __init__(self, means, stds, low, high):
        self.alphas, self.betas, self.locs, self.scales = estimate_beta_parameters(means, stds, low, high)
        self.means = means
        self.stds = stds
        assert (self.alphas < 0).sum() == 0
        assert (self.betas < 0).sum() == 0

    def __iter__(self):
        return iter((self.alphas, self.betas, self.means, self.stds))



from .utils import estimate_beta_parameters

class People:
    def __init__(self, pollution_pref, nuclear_pref, lights_pref, heater_pref,
                 lights_interraction, heaters_interraction):
        # pref is simulated using beta function
        self.pollution = pollution_pref
        self.nuclear = nuclear_pref
        self.lights_pref = lights_pref
        self.heater_pref = heater_pref
        self.lights_interraction = lights_interraction
        self.heaters_interraction = heaters_interraction

    def __str__(self):
        pass

    def to_json(self):
        pass


""" This module contains all simulation functions for the entire city. Those
are called at each tick.
"""
from .scores import compute_scores


class NotEnoughEnergy(Exception):
    def __init__(self, bought, needed):
        msg = f"You bought {bought} energy but you need {needed}"
        super(NotEnoughEnergy, self).__init__(msg)


class SmartCity:
    def __init__(self, people, energies, lights, heaters):
        self.people = people
        self.energies = energies
        self.lights = lights
        self.heaters = heaters

        self.needded_energy = 50
        self.stored = 0

    def step(self):
        """ Compute (and returns) the score for the current step, then update
        the city's data to the next step.
        """
        if self.stored < self.needded_energy:
            raise NotEnoughEnergy(self.energies.bought_amounts.sum(), self.needded_energy)
        scores = compute_scores(people=self.people, energies=self.energies,
                                lights=self.lights, heaters=self.heaters)
        self.energies_step()
        return scores

    def energies_step(self):
        assert self.stored >= self.needded_energy
        new_cost = self.energies.new_cost(self.energies.costs, self.energies.bought_amounts)
        new_amount = self.energies.new_amount(self.energies.amounts, self.energies.bought_amounts)

        self.energies.costs = new_cost
        self.energies.amounts = new_amount

    def light_up(self, lights_on):
        pass

    def heat_up(self, heats, temperatures):
        pass

    def buy_energies(self, amounts):
        cost, quantity = self.energies.buy(amounts)
        self.stored += quantity
        return cost



import numpy as np

from .energy import Energies
from .people import People
from .utils import BetaDistributions


def build_all(configs):
    energies = build_energies(configs["energies"])
    people = build_people(configs["people"], configs["lights"]
                          ["nb_lights"], configs["heaters"]["nb_heaters"])
    lights = build_lights(configs["lights"])
    heaters = build_heaters(configs["heaters"])

    return {
        "energies": energies,
        "people": people,
        "lights": lights,
        "heaters": heaters
    }


def build_energies(config):
    costs = np.array([e["cost"] for e in config["energies"]], dtype=np.float64)
    amounts = np.array([e["amount"]
                        for e in config["energies"]], dtype=np.float64)
    pollution_factors = np.array([e["pollution_factor"]
                                  for e in config["energies"]], dtype=np.float64)

    def new_cost(costs, bought):
        return costs

    def new_amount(amount, bought):
        return amount

    return Energies(
        initial_costs=costs,
        initial_amounts=amounts,
        pollution_factors=pollution_factors,
        new_cost=new_cost,
        new_amount=new_amount
    )


def mean_var(low, high, size, mini, maxi):
    means = np.random.uniform(low=low, high=high, size=size)
    variances = np.random.uniform(low=0.1 * means, high=0.3 * means, size=size)
    return BetaDistributions(means, variances, mini, maxi)


distributions = {"mean_var": mean_var,
                 }


def draw_from_dist(distribution_name, **kwparameters):
    return distributions[distribution_name](**kwparameters)


def random_β_distribution(α_range, β_range):
    α, β = np.random.rand(2)
    α = α * α_range[1] + α_range[0]
    β = β * β_range[1] + β_range[0]
    return α, β


interractions = {"random": 
                    lambda peo, ligh, heat: (np.random.choice([0, 1], size=(peo, ligh), p=[0.8, 0.2]),
                                             np.random.choice([0, 1], size=(peo, heat), p=[0.9, 0.1])),
                }

def draw_interraction(nb_people, nb_lights, nb_heaters, style="random"):
    return interractions[style](nb_people, nb_lights, nb_heaters)


def build_people(config, nb_lights, nb_heaters):
    pollution_pref_params = config["pollution_pref"]
    nuclear_pref_params = config["nuclear_pref"]
    lights_pref_params = config["lights_pref"]
    heat_pref_params = config["heat_pref"]
    nb_people = config["nb_people"]

    print("Pollution")
    pollution_pref = draw_from_dist(size=nb_people, **pollution_pref_params)
    print("Nuclear")
    nuclear_pref = draw_from_dist(size=nb_people, **nuclear_pref_params)
    print("Lights")
    lights_pref = draw_from_dist(size=nb_people, **lights_pref_params)
    print("Heat")
    heat_pref = draw_from_dist(size=nb_people, **heat_pref_params)
    light_inter, heat_inter = draw_interraction(
        nb_people, nb_lights, nb_heaters, config["interraction"])

    return People(
        pollution_pref=pollution_pref,
        nuclear_pref=nuclear_pref,
        lights_pref=lights_pref,
        heater_pref=heat_pref,
        lights_interraction=light_inter,
        heaters_interraction=heat_inter
    )


def build_lights(config):
    return np.ones(config["nb_lights"])


def build_heaters(config):
    heaters = np.random.randn(config["nb_heaters"])
    heaters = (config["mean_temperature"] + heaters) * config["std_temperature"]
    return heaters
